from datetime import date as date_typing, datetime
from typing import Dict, Optional, Union
from urllib.parse import urljoin, urlparse

from aiohttp import ClientSession
from aiohttp.client_reqrep import ClientResponse
from bs4 import BeautifulSoup

from .exceptions import CeiCrawlerUnableToLoginException
from .models import Broker, BrokerAccount


class CeiClient:
    __is_logged__ = False
    LOGIN_URL = "https://ceiapp.b3.com.br/CEI_Responsivo/login.aspx"
    _HEADERS_ORIGIN_URL = "https://cei.b3.com.br"
    WRONG_CREDENTIALS_TEXTS = (
        "Usuário/Senha/Código de verificação inválido(a)",
        "Sua sessão expirou. Favor efetuar um novo acesso.",
        "não há dados disponíveis para serem consultados.",
    )
    CEI_SERVER_ERROR_TEXT = "Desculpe-nos pelo transtorno. Ocorreu um erro inesperado."

    def __init__(
        self,
        username: str,
        password: str,
        session: ClientSession,
        base_url: str,
    ) -> None:
        self.username = username
        self.password = password
        self.session = session
        self.base_url = base_url

    async def _get_referer(self) -> str:
        return urljoin(self._HEADERS_ORIGIN_URL, urlparse(self.base_url).path)

    async def _get_headers(self) -> Dict[str, str]:
        return {
            "content-type": "application/x-www-form-urlencoded; charset=UTF-8",
            "Referer": await self._get_referer(),
            "Origin": self._HEADERS_ORIGIN_URL,
            "User-Agent": (
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) "
                "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 "
                "Safari/537.36"
            ),
        }

    async def login(self) -> bool:
        async with self.session.get(self.LOGIN_URL) as response:
            html = BeautifulSoup(await response.text(), "html.parser")
            try:
                view_state = html.find(id="__VIEWSTATE")["value"]
                viewstate_generator = html.find(id="__VIEWSTATEGENERATOR")["value"]
                event_validation = html.find(id="__EVENTVALIDATION")["value"]
            except TypeError as error:
                raise CeiCrawlerUnableToLoginException(repr(error)) from error

        payload = {
            "ctl00$ContentPlaceHolder1$smLoad": (
                "ctl00$ContentPlaceHolder1$UpdatePanel1|ctl00$ContentPlaceHold"
                "er1$btnLogar"
            ),
            "__EVENTTARGET": "",
            "__EVENTARGUMENT": "",
            "__VIEWSTATEGENERATOR": viewstate_generator,
            "__EVENTVALIDATION": event_validation,
            "__VIEWSTATE": view_state,
            "ctl00$ContentPlaceHolder1$txtLogin": self.username,
            "ctl00$ContentPlaceHolder1$txtSenha": self.password,
            "__ASYNCPOST": True,
            "ctl00$ContentPlaceHolder1$btnLogar": "Entrar",
        }

        headers = {
            "content-type": "application/x-www-form-urlencoded; charset=UTF-8",
            "Referer": "https://cei.b3.com.br/CEI_Responsivo/login.aspx",
            "Origin": self._HEADERS_ORIGIN_URL,
            "Host": "cei.b3.com.br",
            "User-Agent": (
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_5) "
                "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 "
                "Safari/537.36"
            ),
        }

        async with self.session.post(
            self.LOGIN_URL,
            data=payload,
            headers=headers,
        ) as response:

            html = await response.text()
            if any(text in html for text in self.WRONG_CREDENTIALS_TEXTS):
                raise CeiCrawlerUnableToLoginException("Usuário/Senha inválido(a)")

            if self.CEI_SERVER_ERROR_TEXT in html:
                raise CeiCrawlerUnableToLoginException(
                    "Parece que o site do CEI está fora do ar. "
                    f"Resposta do servidor: {self.CEI_SERVER_ERROR_TEXT}"
                )

            self.__is_logged__ = True
            return self.__is_logged__

    async def get_brokers(self) -> ClientResponse:
        if not self.__is_logged__:
            await self.login()
        return await self.session.get(self.base_url)

    async def get_broker_accounts(
        self, payload: Dict[str, Union[str, bool]]
    ) -> ClientResponse:
        if not self.__is_logged__:
            await self.login()
        return await self.session.post(
            self.base_url, data=payload, headers=await self._get_headers()
        )

    async def get_broker_account_portfolio_assets_extract(
        self,
        broker: Broker,
        account: BrokerAccount,
        start_date: Optional[date_typing] = None,
        end_date: Optional[date_typing] = None,
    ) -> ClientResponse:
        if not self.__is_logged__:
            await self.login()

        start_date_str = (
            start_date.strftime("%d/%m/%Y")
            if start_date is not None
            else broker.parse_extra_data.start_date
        )
        end_date_str = (
            end_date.strftime("%d/%m/%Y")
            if end_date is not None
            else broker.parse_extra_data.end_date
        )

        payload = {
            "ctl00$ContentPlaceHolder1$ToolkitScriptManager1": (
                "ctl00$ContentPlaceHolder1$updFiltro|ctl00$ContentPlaceHolder1"
                "$btnConsultar"
            ),
            "__EVENTTARGET": "",
            "__VIEWSTATE": account.parse_extra_data.view_state,
            "__VIEWSTATEGENERATOR": account.parse_extra_data.view_state_generator,
            "__EVENTVALIDATION": account.parse_extra_data.event_validation,
            "ctl00$ContentPlaceHolder1$ddlAgentes": broker.value,
            "ctl00$ContentPlaceHolder1$ddlContas": account.id,
            "ctl00$ContentPlaceHolder1$txtDataDeBolsa": start_date_str,
            "ctl00$ContentPlaceHolder1$txtDataAteBolsa": end_date_str,
            "ctl00$ContentPlaceHolder1$btnConsultar": "Consultar",
            "__ASYNCPOST": True,
        }

        return await self.session.post(
            self.base_url, data=payload, headers=await self._get_headers()
        )

    @staticmethod
    async def _get_date_str(broker: Broker, date: Optional[date_typing]) -> str:
        if (
            date
            and datetime.strptime(broker.parse_extra_data.start_date, "%d/%m/%Y").date()
            <= date
            <= datetime.strptime(broker.parse_extra_data.end_date, "%d/%m/%Y").date()
        ):
            # By some reason, CEI will return some "nonsense" data if
            # we pass a date out of range
            return date.strftime("%d/%m/%Y")
        return broker.parse_extra_data.end_date

    async def get_passive_incomes_extract(
        self, broker: Broker, date: Optional[date_typing] = None
    ) -> ClientResponse:
        if not self.__is_logged__:
            await self.login()

        account = broker.accounts[0]  # accounts[0] it's relative to all accounts
        payload = {
            "ctl00$ContentPlaceHolder1$ToolkitScriptManager1": (
                "ctl00$ContentPlaceHolder1$updFiltro|ctl00$ContentPlaceHolder1"
                "$btnConsultar"
            ),
            "__EVENTTARGET": "",
            "__VIEWSTATE": account.parse_extra_data.view_state,
            "__VIEWSTATEGENERATOR": account.parse_extra_data.view_state_generator,
            "__EVENTVALIDATION": account.parse_extra_data.event_validation,
            "ctl00$ContentPlaceHolder1$ddlAgentes": broker.value,
            "ctl00$ContentPlaceHolder1$ddlContas": account.id,
            "ctl00$ContentPlaceHolder1$txtData": await self._get_date_str(
                broker=broker, date=date
            ),
            "ctl00$ContentPlaceHolder1$btnConsultar": "Consultar",
            "__ASYNCPOST": True,
        }

        return await self.session.post(
            self.base_url, data=payload, headers=await self._get_headers()
        )



import re
from typing import Any, AsyncGenerator, Dict, List, Tuple, Union

from aiohttp.client_reqrep import ClientResponse
from bs4 import BeautifulSoup
from bs4.element import ResultSet, Tag

from .models import (
    Broker,
    BrokerAccount,
    BrokerAccountParseExtraData,
    AssetExtract,
    BrokerParseExtraData,
    PassiveIncome,
)


class _BrokerResponseBase:
    BROKERS_SELECT_ID = "ctl00_ContentPlaceHolder1_ddlAgentes"
    DEFAULT_INVALID_BROKER_VALUE = "-1"

    def __init__(self, response: ClientResponse) -> None:
        self.response = response

    @staticmethod
    async def _get_date_input_ids() -> Tuple[str, str]:
        raise NotImplementedError()

    async def data(self) -> List[Broker]:
        html = await self.response.text()
        return await self._parse_brokers(html)

    async def _parse_brokers(self, html: str) -> List[Broker]:
        soup = BeautifulSoup(html, "html.parser")
        brokers_select = soup.find("select", id=self.BROKERS_SELECT_ID)
        brokers_option = brokers_select.find_all("option")
        start_date_id, end_date_id = await self._get_date_input_ids()
        start_date = soup.find(id=start_date_id).text
        end_date = soup.find(id=end_date_id).text

        return [
            Broker(
                name=broker_option.text,
                value=broker_option["value"],
                parse_extra_data=BrokerParseExtraData(start_date, end_date),
            )
            for broker_option in brokers_option
            if broker_option["value"] != self.DEFAULT_INVALID_BROKER_VALUE
        ]


class BrokerAssetsResponse(_BrokerResponseBase):
    @staticmethod
    async def _get_date_input_ids() -> Tuple[str, str]:
        start_date_id = "ctl00_ContentPlaceHolder1_lblPeriodoInicialBolsa"
        end_date_id = "ctl00_ContentPlaceHolder1_lblPeriodoFinalBolsa"
        return start_date_id, end_date_id


class BrokerPassiveIncomesResponse(_BrokerResponseBase):
    @staticmethod
    async def _get_date_input_ids() -> Tuple[str, str]:
        start_date_id = "ctl00_ContentPlaceHolder1_lblPeriodoInicial"
        end_date_id = "ctl00_ContentPlaceHolder1_lblPeriodoFinal"
        return start_date_id, end_date_id


class BrokerAccountResponse:
    ACCOUNT_SELECT_ID = "ctl00_ContentPlaceHolder1_ddlContas"
    BROKERS_SELECT_ID = "ctl00_ContentPlaceHolder1_ddlAgentes"

    def __init__(self, response: ClientResponse, broker: Broker) -> None:
        self.response = response
        self.broker = broker

    async def data(self) -> Broker:
        html = await self.response.text()

        self.broker.accounts = await self._parse_accounts(html)

        return self.broker

    async def _parse_accounts(self, html: str) -> List[BrokerAccount]:
        soup = BeautifulSoup(html, "html.parser")
        brokers_select = soup.find("select", id=self.ACCOUNT_SELECT_ID)
        if not brokers_select:
            return []

        brokers_option = brokers_select.find_all("option")
        view_state = soup.find(id="__VIEWSTATE")["value"]
        view_state_generator = soup.find(id="__VIEWSTATEGENERATOR")["value"]
        event_validation = soup.find(id="__EVENTVALIDATION")["value"]

        parse_extra_data = BrokerAccountParseExtraData(
            view_state=view_state,
            view_state_generator=view_state_generator,
            event_validation=event_validation,
        )

        return [
            BrokerAccount(id=broker_option["value"], parse_extra_data=parse_extra_data)
            for broker_option in brokers_option
        ]


class BrokerAccountAssetExtractResponse:
    ASSETS_TABLE_ID = (
        "ctl00_ContentPlaceHolder1_rptAgenteBolsa_ctl00_rptContaBolsa_ctl00_"
        "pnAtivosNegociados"
    )

    def __init__(self, response: ClientResponse, broker_value: str) -> None:
        self.response = response
        self.broker_value = broker_value

    async def data(
        self, as_dict: bool = False
    ) -> Union[List[AssetExtract], List[Dict[str, Any]]]:
        html = await self.response.text()
        return await self._parse_assets_extract(html, as_dict=as_dict)

    async def _parse_assets_extract(
        self, html: str, as_dict: bool
    ) -> Union[List[AssetExtract], List[Dict[str, Any]]]:
        assets_extract = []
        soup = BeautifulSoup(html, "html.parser")
        assets_table = soup.find(id=self.ASSETS_TABLE_ID)

        table_body = assets_table.find("tbody") if assets_table else ResultSet(None)
        rows = table_body.find_all("tr") if table_body else []
        for row in rows:
            (
                operation_date,
                action,
                market_type,
                _,
                raw_negotiation_code,
                asset_specification,
                unit_amount,
                unit_price,
                total_price,
                quotation_factor,
            ) = row.find_all("td")
            asset_extract = await AssetExtract.create_from_response_fields(
                operation_date=operation_date.get_text(strip=True),
                action=action.get_text(strip=True),
                market_type=market_type.get_text(strip=True),
                raw_negotiation_code=raw_negotiation_code.get_text(strip=True),
                asset_specification=asset_specification.get_text(strip=True),
                unit_amount=unit_amount.get_text(strip=True),
                unit_price=unit_price.get_text(strip=True),
                total_price=total_price.get_text(strip=True),
                quotation_factor=quotation_factor.get_text(strip=True),
            )
            assets_extract.append(asset_extract.__dict__ if as_dict else asset_extract)

        return assets_extract  # type: ignore


class PassiveIncomesResponse:
    PASSIVE_INCOMES_DOCUMENT_ID = "ctl00_ContentPlaceHolder1_updFiltro"
    BROKER_TITLE_ID_REGEX = (
        r"ctl00_ContentPlaceHolder1_rptAgenteProventos.*lblAgenteProventos"
    )
    UNSUPPORTED_EVENT_TYPE = "Eventos em Ativos Creditado"

    def __init__(self, response: ClientResponse) -> None:
        self.response = response

    async def data(
        self, as_dict: bool = False
    ) -> Union[List[PassiveIncome], List[Dict[str, Any]]]:
        html = await self.response.text()
        return await self._parse_data(html=html, as_dict=as_dict)

    @staticmethod
    async def _parse_passive_incomes(
        table: Tag, event_type: str
    ) -> AsyncGenerator[PassiveIncome, None]:
        for row in table.find_all("tr"):
            (
                raw_negotiation_name,
                asset_specification,
                raw_negotiation_code,
                operation_date,
                income_type,
                unit_amount,
                quotation_factor,
                gross_value,
                net_value,
            ) = row.find_all("td")
            yield await PassiveIncome.create_from_response_fields(
                raw_negotiation_name=raw_negotiation_name.get_text(strip=True),
                asset_specification=asset_specification.get_text(strip=True),
                raw_negotiation_code=raw_negotiation_code.get_text(strip=True),
                operation_date=operation_date.get_text(strip=True),
                income_type=income_type.get_text(strip=True),
                unit_amount=unit_amount.get_text(strip=True),
                quotation_factor=quotation_factor.get_text(strip=True),
                gross_value=gross_value.get_text(strip=True),
                net_value=net_value.get_text(strip=True),
                event_type=event_type,
            )

    @staticmethod
    async def _get_event_types(broker_tag: Tag, method_name: str) -> ResultSet:
        method = getattr(broker_tag, method_name)
        return method("p", class_="title")

    async def _parse_data(
        self, html: str, as_dict: bool
    ) -> Union[List[PassiveIncome], List[Dict[str, Any]]]:
        passive_incomes = []
        soup = BeautifulSoup(html, "html.parser")
        document = soup.find(id=self.PASSIVE_INCOMES_DOCUMENT_ID)
        brokers = document.find_all(id=re.compile(self.BROKER_TITLE_ID_REGEX))
        num_of_brokers = len(brokers)

        for i in range(num_of_brokers):
            if i + 1 == num_of_brokers:  # last broker
                broker_tag = brokers[i]
                method_name = "find_all_next"
            else:
                broker_tag = brokers[i + 1]
                method_name = "find_all_previous"

            for event_type in await self._get_event_types(
                broker_tag=broker_tag, method_name=method_name
            ):
                if event_type.get_text(strip=True) == self.UNSUPPORTED_EVENT_TYPE:
                    continue

                async for passive_income in self._parse_passive_incomes(
                    table=event_type.find_next("tbody") or [],
                    event_type=event_type.get_text(strip=True).split()[-1],
                ):
                    passive_incomes.append(
                        passive_income.__dict__ if as_dict else passive_income
                    )

        return passive_incomes  # type: ignore



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cei_crawler', 'cei_crawler.tests']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp==3.7.4', 'beautifulsoup4==4.10.0']

setup_kwargs = {
    'name': 'python-cei-crawler',
    'version': '0.1.1',
    'description': 'Biblioteca em python para obtenção de seus dados de investimentos na bolsa de valores (B3/CEI).',
    'long_description': '# python-cei-crawler\nBiblioteca em python para obtenção de seus dados de investimentos na bolsa de valores (B3/CEI).\n\nEsse projeto é altamente influenciado por [bolsa](https://github.com/gicornachini/bolsa). De fato, eu apenas simplifiquei, adicionei e estendi algumas funcionalidades.\n\n# Requisitos\n - Python 3.8.x\n\n## Instalação\n```\n$ pip install python-cei-crawler\n```\n\n## Utilização\n```python\nimport asyncio\n\nfrom cei_crawler import CeiCrawler\n\n\nasync def main():\n    crawler = CeiCrawler(username="CPF/CNPJ", password="Sua senha")\n    \n    assets_extract = await crawler.get_assets_extract()\n    print(assets_extract) # seus ativos negociados no CEI\n\n    passive_income_extract = await crawler.get_passive_incomes_extract()\n    print(passive_income_extract) # seus proventos registrados no CEI\n\n    await crawler.close()\n\n\nif __name__ == "__main__":\n    loop = asyncio.get_event_loop()\n    loop.run_until_complete(main())\n\n```\n\n## Funções disponíveis\n\nAtravés da classe de client `CeiCrawler`, você terá acesso as seguintes funções:\n\n### `get_brokers`\nObtém os brokers disponíveis para aquela conta. Retorna uma lista de `Broker` (Ex: XP Inc, Clear, Easynvest...) com uma lista de `BrokerAccount`.\n\n### `get_assets_extract`\nObtém uma lista de ativos filtrados pelo parâmetros passados à função. Retorna uma lista de `AssetExtract`.\n\n#### Parâmetros\n| Atributo | Tipo | Descrição |\n| :-------------: |:-------------:| -----|\n| `brokers` | `Optional[List[Broker]]` | Retorna apenas os ativos destes brokers |\n| `start_date` | `Optional[date]` | Retorna apenas os ativos com data posterior ou igual a esta |\n| `end_date` | `Optional[date]` | Retorna apenas os ativos com data inferior ou igual a esta |\n| `as_dict` | `bool` | Retorna os ativos como dicionários ao invés de objetos `AssetExtract`. Default: `False` |\n\n### `get_passive_incomes_extract`\nObtém uma lista de rendimentos passivos filtrados pelo parâmetros passados à função. Retorna uma lista de `PassiveIncome`.\n\n#### Parâmetros\n| Atributo | Tipo | Descrição |\n| :-------------: |:-------------:| -----|\n| `date` | `Optional[date]` | Retorna apenas os ativos com data posterior ou igual a esta. Por favor, note que o CEI só aceita datas iguais ou 5 dias anteriores ao dia de hoje. Se uma data fora desse intervalo for passado, a consulta será feita com a data de hoje. |\n| `as_dict` | `bool` | Retorna os ativos como dicionários ao invés de objetos `PassiveIncome`. Default: `False` |\n\n## Models\n\n#### Broker\nModel responsável pelos dados do broker.\n\n| Atributo        | Tipo           | Descrição  |\n| :-------------: |:-------------:| -----|\n| `value`      | `str` | Identificador da corretora na B3. |\n| `name`      | `str`      |   Nome do broker na B3. |\n| `accounts` | `List[BrokerAccount]`      |    Lista de contas no broker. |\n\n\n#### BrokerAccount\nModel responsável pelos dados da conta no broker.\n\n| Atributo        | Tipo           | Descrição  |\n| :-------------: |:-------------:| -----|\n| `id`      | `str` | Número da conta no broker. |\n\n\n#### AssetExtract\nModel responsável pelos dados do ativo.\n\n| Atributo        | Tipo           | Descrição  |\n| :-------------: |:-------------:| -----|\n| `operation_date`      | `datetime` | Data de operação do ativo. |\n| `action`      | `AssetExtractAction`      |   Identificador do tipo de operação compra/venda. |\n| `market_type` | `AssetExtractMarketType`      |   Tipo de mercado. |\n| `raw_negotiation_code` | `str`      |    Código de negociação. |\n| `asset_specification` | `str`      |    Especificação do ativo no CEI. |\n| `unit_amount` | `int`      |    Quantidade de ativo. |\n| `unit_price` | `Decimal`      |    Valor unitário do ativo. |\n| `total_price` | `Decimal`      |    Valor total do ativo. |\n| `quotation_factor` | `int`      |    Fator de cotação. |\n\n#### PassiveIncome\nModel responsável pelos dados de rendimento passivo.\n\n| Atributo        | Tipo           | Descrição  |\n| :-------------: |:-------------:| -----|\n| `operation_date`      | `datetime` | Data do evento. |\n| `income_type`      | `PassiveIncomeType`      |   Tipo de provento (Dividendo, JCP...). |\n| `event_type`      | `PassiveIncomeEventType`      |   Tipo de evento do provento (provisionado, creditado...). |\n| `raw_negotiation_name` | `str`      |    Nome do ativo. |\n| `raw_negotiation_code` | `str`      |    Código de negociação. |\n| `asset_specification` | `str`      |    Especificação do ativo no CEI. |\n| `unit_amount` | `int`      |    Quantidade de ativo. |\n| `gross_value` | `Decimal`      |    Valor bruto do provento. |\n| `net_value` | `Decimal`      |    Valor líquido do provendo. |\n| `quotation_factor` | `int`      |    Fator de cotação. |',
    'author': 'MuriloScarpaSitonio',
    'author_email': 'muriloscarpa@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/MuriloScarpaSitonio/python-cei-crawler',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
